@@ -1,10 +1,15 @@
+from typing import Optional, Union
+
+from decimal import Decimal
 import docassemble.base.functions
 from docassemble.base.util import (
     defined,
     value,
     showifdef,
     space_to_underscore,
+    action_button_html,
     Address,
+    word,
 )
 import re
 
@@ -18,6 +23,7 @@ __all__ = [
     "number_to_letter",
     "collapse_template",
     "tabbed_templates_html",
+    "review_widget",
     "sum_if_defined",
     "add_records",
     "output_checkbox",
@@ -33,12 +39,14 @@ class shortenMe:
 
 
 # The following three functions are from Quinten
-# format a float number to a whole number with thousands separator
-def thousands(num: float, show_decimals=False) -> str:
+def thousands(num: Union[float, str, Decimal], show_decimals=False) -> str:
     """
     Return a whole number formatted with thousands separator.
     Optionally, format with 2 decimal points (for a PDF form with the
     currency symbol already present in the form)
+
+    If `trucate`, will call `int(num)`, truncating the decimals instead of
+    rounding to the closest int.
     """
     try:
         if show_decimals:
@@ -46,46 +54,51 @@ def thousands(num: float, show_decimals=False) -> str:
         else:
             return f"{int(num):,}"
     except:
-        return num
+        return str(num)
 
 
-# Format a phone number so you can click on it to open in your phone dialer
-def tel(phone_number):
+def tel(phone_number) -> str:
+    """Format a phone number so you can click on it to open in your phone dialer"""
     return '<a href="tel:' + str(phone_number) + '">' + str(phone_number) + "</a>"
 
 
-# Display an icon on the screen.
-def fa_icon(icon, color="primary", color_css=None, size="sm"):
+def fa_icon(
+    icon: str, color="primary", color_css=None, size="sm", fa_class="fa-solid"
+) -> str:
     """
     Return HTML for a font-awesome icon of the specified size and color. You can reference
     a CSS variable (such as Bootstrap theme color) or a true CSS color reference, such as 'blue' or
     '#DDDDDD'. Defaults to Bootstrap theme color "primary".
+
+    Sizes can be '2xs', 'xs', 'sm', 'md' (or None), 'lg', 'xl', '2xl'.
     """
-    if not color and not color_css:
+    if not size or size == "md":
+        size_str = ""
+    else:
+        size_str = " fa-" + size
+    if not size and not color and not color_css:
         return ":" + icon + ":"  # Default to letting Docassemble handle it
     elif color_css:
         return (
-            '<i class="fa fa-'
+            f'<i class="{fa_class} fa-'
             + icon
-            + " fa-"
-            + size
+            + size_str
             + '" style="color:'
             + color_css
             + ';"></i>'
         )
     else:
         return (
-            '<i class="fa fa-'
+            f'<i class="{fa_class} fa-'
             + icon
-            + " fa-"
-            + size
+            + size_str
             + '" style="color:var(--'
             + color
             + ');"></i>'
         )
 
 
-def space(var_name, prefix=" ", suffix=""):
+def space(var_name: str, prefix=" ", suffix="") -> str:
     """If the value as a string is defined, return it prefixed/suffixed. Defaults to prefix
     of a space. Helps build a sentence with less cruft. Equivalent to SPACE function in
     HotDocs."""
@@ -101,7 +114,9 @@ def space(var_name, prefix=" ", suffix=""):
         return ""
 
 
-def yes_no_unknown(var_name, condition, unknown="Unknown", placeholder=0):
+def yes_no_unknown(
+    var_name: str, condition: Optional[bool], unknown="Unknown", placeholder=0
+):
     """Return 'unknown' if the value is None rather than False. Helper for PDF filling with
     yesnomaybe fields"""
     if condition:
@@ -112,7 +127,7 @@ def yes_no_unknown(var_name, condition, unknown="Unknown", placeholder=0):
         return placeholder
 
 
-def number_to_letter(n):
+def number_to_letter(n: Optional[int]) -> str:
     """Returns a capital letter representing ordinal position. E.g., 1=A, 2=B, etc. Appends letters
     once you reach 26 in a way compatible with Excel/Google Sheets column naming conventions. 27=AA, 28=AB...
     """
@@ -125,9 +140,19 @@ def number_to_letter(n):
     return string
 
 
-def collapse_template(template, classname=None):
+def collapse_template(
+    template,
+    classname=None,
+    closed_icon="caret-right",
+    open_icon="caret-down",
+):
     """
     Insert HTML for a Bootstrap "collapse" div.
+
+    Optionally, you can specify a custom icon to override the defaults:
+
+    The default icons are "🞂" which displays when the text is collapsed (`closed_icon`) and
+    "▼" which displays when the text is open (`open_icon`).
     """
     if not template.subject_as_html(trim=True) and not template.content_as_html():
         return ""
@@ -136,16 +161,10 @@ def collapse_template(template, classname=None):
     else:
         classname = " " + classname.strip()
     the_id = re.sub(r"[^A-Za-z0-9]", "", template.instanceName)
-    return """\
-<a class="collapsed" data-bs-toggle="collapse" href="#{}" role="button" aria-expanded="false" aria-controls="collapseExample"><span class="pdcaretopen"><i class="fas fa-caret-down"></i></span><span class="pdcaretclosed"><i class="fas fa-caret-right"></i></span> {}</a>
-<div class="collapse" id="{}"><div class="card card-body{} pb-1">{}</div></div>\
-""".format(
-        the_id,
-        template.subject_as_html(trim=True),
-        the_id,
-        classname,
-        template.content_as_html(),
-    )
+    return f"""\
+<a class="collapsed" data-bs-toggle="collapse" href="#{ the_id }" role="button" aria-expanded="false" aria-controls="{ the_id }"><span class="pdcaretopen">{ fa_icon(open_icon) }</span><span class="pdcaretclosed">{ fa_icon(closed_icon) }</span> { template.subject_as_html(trim=True) }</a>
+<div class="collapse" id="{ the_id }"><div class="card card-body pb-1{ classname }">{ template.content_as_html() }</div></div>\
+"""
 
 
 def tabbed_templates_html(tab_group_name: str, *pargs) -> str:
@@ -179,6 +198,66 @@ def tabbed_templates_html(tab_group_name: str, *pargs) -> str:
     return tabs + tab_content
 
 
+def review_widget(
+    *,
+    up_action: str,
+    down_action: str,
+    review_action: Optional[str] = None,
+    thumbs_display: str = "Did we help you?",
+    review_display: str = "Thanks! Let us know what we could do better",
+    submit_review_button: str = "Add your review",
+    post_review_display: str = "Thank you for your review!",
+) -> str:
+    """
+    A widget that allows people to give a quick review (thumbs up and down, with an optional text
+    component) in the middle of an interview without triggering a page reload.
+
+    If `review_action` is provided, once you press either of the thumbs, a text input screen with
+    a submit button appears, and once the text review is submitted (or after the thumbs, if no
+    `review_action` was provided), a final "thank you" messsage is displayed.
+
+    @param up_action {string} - the variable name of an event to be executed on the server if the
+        thumbs up is pressed
+    @param down_action {string} - the variable name of an event to be executed on the server i
+        the thumbs down is pressed
+    @param (optional) review_action {string} - the variable name of an event to be execute on the
+        server when someone submits their text review
+    @param (optional) thumbs_display {string} - text displayed to user describing the thumbs
+    @param (optional) review_display {string} - text displayed to user describing the text input
+    @param (optional) submit_review_button {string} - text on the button to submit their text review
+    @param (optional) post_review_display {string} - text displayed to user after review is
+        submitted
+    """
+    js_thumbs_up = f"javascript:altoolbox_thumbs_up_send('{up_action}', {'true' if review_action else 'false'})"
+    js_thumbs_down = f"javascript:altoolbox_thumbs_down_send('{down_action}', {'true' if review_action else 'false'})"
+    widget = f"""
+<div class="card" style="width: 20rem;">
+  <div class="card-body">
+    <p class="al-thumbs-widget">{word(thumbs_display)}</p>
+    <a href="{js_thumbs_up}" id="al-thumbs-widget-up"
+        class="btn btn-md btn-info al-thumbs-widget" aria-label="{word('Thumbs up')}">{fa_icon('thumbs-up', size='md')}</a>
+    <a href="{js_thumbs_down}" id="al-thumbs-widget-down"
+        class="btn btn-md btn-info al-thumbs-widget" aria-label="{word('Thumbs down')}">{fa_icon('thumbs-down', size='md')}</a>
+    """
+    if review_action:
+        review_area_id = review_action + "_area_id"
+        js_review = (
+            f"javascript:altoolbox_review_send('{review_action}', '{review_area_id}')"
+        )
+        widget += f"""
+      <p class="al-review-text al-hidden">{word(review_display)}</p>
+      <textarea class="datextarea al-review-text al-hidden" id="{review_area_id}"
+          alt="{word('Write your review here')}" rows="4"></textarea>
+      <br class="al-review-text">
+      {action_button_html(js_review, label=word(submit_review_button), size='md', classname='al-review-text al-hidden')} 
+        """
+    widget += f"""
+  <p class="al-after-review al-hidden">{word(post_review_display)}</p>
+  </div>
+</div>"""
+    return widget
+
+
 def sum_if_defined(*pargs):
     """Lets you add up the value of variables that are not in a list"""
     total = 0
